#include "Mace.h"

#ifdef OLD_FLATTEN
extern int Dp_sn;  /* owned by generate.c */
#endif

/*************
 *
 *   kludgey_e_subsume(c, d)
 *
 *   Does c subsume d (mod equality flipping)?  This works by
 *   inserting flips into (a copy of) d then doing an ordinary
 *   subsumption check.
 *
 *************/

int kludgey_e_subsume(struct clause *c,
		      struct clause *d)
{
  struct clause *d1;
  struct literal *l1, *l2;
  struct term *t;
  int rc;
    
  d1 = cl_copy(d);
  for (l1 = d1->first_lit; l1; l1 = l1->next_lit) {
    if (is_eq(l1->atom->sym_num)) {
      l2 = get_literal();
      l2->next_lit = d1->first_lit;
      d1->first_lit = l2;
      l2->sign = l1->sign;
      l2->atom = copy_term(l1->atom);
      t = l2->atom->farg->argval;
      l2->atom->farg->argval = l2->atom->farg->narg->argval;
      l2->atom->farg->narg->argval = t;
    }
  }
  rc = subsume(c, d1);
  cl_del_non(d1);
  return(rc);
}  /* kludgey_e_subsume */

/*************
 *
 *   check_for_bad_things()
 *
 *************/

void check_for_bad_things(struct clause *c)
{
  struct literal *lit;
  for (lit = c->first_lit; lit; lit = lit->next_lit) {
    if (answer_lit(lit))
      abend("answer literals not allowed in MACE input");
    if (lit->atom->varnum == EVALUABLE)
      abend("evaluable literals not allowed in MACE input");
  }
}  /* check_for_bad_things */

/*************
 *
 *   int_term(t)
 *
 *   Is t a constant with a symbol representing a positive integer?
 *
 *************/

int int_term(struct term *t)
{
  int i;
  return(t->type == NAME && str_int(sn_to_str(t->sym_num),&i) && i>=0);
}  /* int_term */

/*************
 *
 *   domain_element(t)
 *
 *   Is t a domain element?
 *
 *************/

int domain_element(struct term *t)
{
  return(Flags[DP_INT_DOMAIN].val && int_term(t));
}  /* int_term */

/*************
 *
 *    build_binary_term
 *
 *************/

static struct term *build_binary_term(int sn,
				      struct term *t1,
				      struct term *t2)
{
  struct term *t;
  struct rel *r1, *r2;

  t = get_term(); t->type = COMPLEX; t->sym_num = sn;
  r1 = get_rel(); r1->argval = t1;
  r2 = get_rel(); r2->argval = t2;
  t->farg = r1; r1->narg = r2;
  return(t);
}  /* build_binary_term */

/*************
 *
 *   replace_term()
 *
 *************/

static struct term *replace_term(struct term *t,
				 struct term *ct,
				 struct term *vt)
{
  if (term_ident(t, ct))
    return(copy_term(vt));
  else {
    struct rel *r;
    for (r = t->farg; r; r= r->narg)
      r->argval = replace_term(r->argval, ct, vt);
    return(t);
  }
}  /* replace_term */

/*************
 *
 *   process_negative_equalities()
 *
 *   Simplify literals x!=t, where x does not occur in t, by,
 *   in effect, rewolving with x=x.
 *
 *************/

void process_negative_equalities(struct clause *c)
{
  struct literal *curr, *prev, *l;
  struct term *t1, *t2, *v, *t;

  t = NULL;  /* to quiet -Wall */

  curr = c->first_lit;
  prev = NULL;
  while (curr) {
    if (is_eq(curr->atom->sym_num) && !curr->sign) {
      t1 = curr->atom->farg->argval;
      t2 = curr->atom->farg->narg->argval;
      v = NULL;
      if (t1->type == VARIABLE && !occurs_in(t1, t2)) {
	v = t1; t = t2;
      }
      else if (t2->type == VARIABLE && !occurs_in(t2, t1)) {
	v = t2; t = t1;
      }

      if (v) {
	if (prev)
	  prev->next_lit = curr->next_lit;
	else
	  c->first_lit = curr->next_lit;
	for (l = c->first_lit; l; l = l->next_lit)
	  l->atom = replace_term(l->atom, v, t);
      }
      else
	prev = curr;
    }
    else
      prev = curr; 

    curr = curr->next_lit;
  }
}  /* process_negative_equalities */

/****************************************************************************/
/****************************************************************************/

#ifdef OLD_FLATTEN

/*************
 *
 *   dp_p_term()
 *
 *************/

void dp_p_term(struct term *t)
{
  if (t->type == VARIABLE)
    printf("v%d", t->varnum);
  else {
    struct rel *r;
    printf("%s", sn_to_str(t->sym_num));
	
    for (r = t->farg; r; r = r->narg) {
      printf(" ");
      dp_p_term(r->argval);
    }
  }
}  /* dp_p_term */

/*************
 *
 *   dp_p_clause()
 *
 *************/

void dp_p_clause(struct clause *c)
{
  struct literal *lit;
  struct term *atom, *t1;
  struct rel *r;

  for (lit = c->first_lit; lit; lit = lit->next_lit) {
    if (lit != c->first_lit)
      printf("   ");
    if (!lit->sign)
      printf("-");
    atom = lit->atom;
    if (atom->sym_num == Dp_sn) {
      t1 = atom->farg->argval;
      printf("%s", sn_to_str(t1->sym_num));
      for (r = t1->farg; r; r = r->narg) {
	printf(" ");
	dp_p_term(r->argval);
      }
      printf(" ");
      dp_p_term(atom->farg->narg->argval);
    }
    else
      dp_p_term(atom);
  }
  printf(" .\n");
}  /* dp_p_clause */

/*************
 *
 *   occurrences_cl()
 *
 *   Return the number of occurrnces of term t in clause c.
 *
 *************/

static int occurrences_cl(struct term *t,
			  struct clause *c)
{
  struct literal *l;
  int n = 0;

  for (l = c->first_lit; l; l = l->next_lit)
    n += occurrences(t, l->atom);
  return(n);
}  /* occurrences_cl */

/*************
 *
 *   clause_to_pair()
 *
 *   This routine does the following type of transformation:
 *   a!=x | b!=y | x=y
 *   is made into the pair of clauses
 *   a!=x | b=x   and   a=x | b!=x.
 *   There can be other literals, and a and b need not be constants.
 *
 *   If this can be done, two new clauses are created, and "returned".
 *
 *************/

int clause_to_pair(struct clause *c,
		   struct clause **dp,
		   struct clause **ep)
{
  struct clause *d, *e;
  struct literal *curr, *prev;
  struct term *t;
  int lits[MAX_VARS];  /* lits[i] is index of Dp_eq literal with right side var i */
  int v1, v2, l1, l2, i;

  prev = NULL;  /* to quiet -Wall */

  for (i = 0; i < MAX_VARS; i++)
    lits[i] = -1;

  /* Assume all Dp_eq lits first, and in Dp_lit, var is second arg. */

  for (curr = c->first_lit, i = 0;
       curr;
       prev = curr, curr = curr->next_lit, i++) {
    if (curr->atom->sym_num == Dp_sn) {
      t = curr->atom->farg->narg->argval;
      if (occurrences_cl(t, c) == 2)
	lits[t->varnum] = i;
    }
    else if (is_eq(curr->atom->sym_num) && curr->sign) {
      struct term *a1 = curr->atom->farg->argval;
      struct term *a2 = curr->atom->farg->narg->argval;

      if (a1->type == VARIABLE && a2->type == VARIABLE) {
	v1 = a1->varnum;
	v2 = a2->varnum;
	if (lits[v1] >= 0 && lits[v2] >= 0) {
	  l1 = lits[v1]; l2 = lits[v2];
	  prev->next_lit = curr->next_lit;
		    
	  for (curr = c->first_lit, i = 0; i < l1; curr = curr->next_lit, i++);
	  curr->atom->farg->narg->argval->varnum = v1;
	  for (curr = c->first_lit, i = 0; i < l2; curr = curr->next_lit, i++);
	  curr->atom->farg->narg->argval->varnum = v1;
		    
	  d = cl_copy(c);
	  e = cl_copy(c);
	  for (curr = d->first_lit, i = 0; i < l1; curr = curr->next_lit, i++);
	  curr->sign = 1;
	  for (curr = e->first_lit, i = 0; i < l2; curr = curr->next_lit, i++);
	  curr->sign = 1;
	  *dp = d; *ep = e;
	  return(1);
	}
      }
    }
  }

  return(0);
}  /* clause_to_pair */

/*************
 *
 *   clause_to_clauses()
 *
 *************/

struct list *clause_to_clauses(struct clause *c)
{
  struct list *l;
  struct clause *d, *e, *f;
  int subsumed;
  l = get_list();
  append_cl(l, c);
  d = l->first_cl;
  while (d) {
    if (clause_to_pair(d, &e, &f)) {
      insert_after_cl(d, e);
      insert_after_cl(e, f);
      rem_from_list(d);
      d = e;
    }
    else
      d = d->next_cl;
  }

  /* Now delete any clauses subsumed by previous clauses in the list. */
  for (d = l->first_cl; d; d = d->next_cl) {
    for (e = l->first_cl, subsumed=0; e != d && !subsumed; e = e->next_cl)
      subsumed = subsume(e,d);
    if (subsumed) {
      f = d;
      d = d->prev_cl;
      rem_from_list(f);
    }
  }

  return(l);
}  /* clause_to_clauses */

/*************
 *
 *   check_transformed_clause()
 *
 *   This checks that the transformed clause is equality-equivalent
 *   to the original.
 *
 *************/

void check_transformed_clause(struct clause *c,
			      struct clause *orig)
{
  struct clause *d;
  struct literal *l;

  d = cl_copy(c);
  for (l = d->first_lit; l; l = l->next_lit) {
    if (l->atom->sym_num == Dp_sn)
      l->atom->sym_num = str_to_sn("=", 2);
  }
  process_negative_equalities(d);
  if (!kludgey_e_subsume(d, orig) || !kludgey_e_subsume(orig, d)) {
    printf("%% WARNING: possible error in check_transformed_clause:\n");
    printf("%%    original:    "); p_clause(orig);
    printf("%%    transformed: "); p_clause(d);
  }
}  /* check_transformed_clause */

/*************
 *
 *   lone_variable_process()
 *
 *   If a pos eq has a var on the right that does not occur in
 *   another literal, and the left is not a variable,
 *   then replace the equality symbol with Dp_sn.  This
 *   simplifies the result, and I think it is complete.
 *
 *************/

void lone_variable_process(struct clause *c)
{
  struct literal *l1, *l2;
  struct term *v;
  int ok;

  for (l1 = c->first_lit; l1; l1 = l1->next_lit) {
    if (l1->sign && is_eq(l1->atom->sym_num) &&
	l1->atom->farg->narg->argval->type == VARIABLE &&
	l1->atom->farg->argval->type != VARIABLE) {
      v = l1->atom->farg->narg->argval;
      ok = 1;
      for (l2 = c->first_lit, ok = 1; l2 && ok; l2 = l2->next_lit)
	ok = (l1 == l2 || !occurs_in(v, l2->atom));
      if (ok)
	l1->atom->sym_num = Dp_sn;
    }
  }
}  /* lone_variable_process */

/*************
 *
 *   domain_element_process()
 *
 *   If Flag[DP_INT_DOMAIN].val, then simplify literals "-<int> <var>"
 *
 *************/

void domain_element_process(struct clause *c)
{
  struct literal *curr, *prev, *l;
  struct term *t1, *t2;

  curr = c->first_lit;
  prev = NULL;
  while (curr) {
    if (curr->atom->sym_num == Dp_sn && !curr->sign) {
      t1 = curr->atom->farg->argval;
      t2 = curr->atom->farg->narg->argval;
      if (domain_element(t1) && t2->type == VARIABLE) {
	if (prev)
	  prev->next_lit = curr->next_lit;
	else
	  c->first_lit = curr->next_lit;
	for (l = c->first_lit; l; l = l->next_lit)
	  l->atom = replace_term(l->atom, t2, t1);
      }
      else
	prev = curr;
    }
    else
      prev = curr; 

    curr = curr->next_lit;
  }
}  /* domain_element_process */

/*************
 *
 *   first_nonvar_arg_term()
 *
 *   Return the first subterm (including the term itself) that
 *   is not a variable and not a domain element.  If the argument
 *   `check' is 0, don't check the term---check only the subterms.
 *   (check is intended to be 0 for atoms and for args of
 *   Dp_sn atoms.)
 *
 *************/

static struct term *first_nonvar_arg_term(struct term *t,
					  int check)
{
  struct rel *r;
  struct term *t1;
  int check_args;

  check_args = (t->sym_num != Dp_sn);

  if (check && t->type != VARIABLE)
    return(t);
  else {
    for (r = t->farg; r; r = r->narg) {
      if ((t1 = first_nonvar_arg_term(r->argval, check_args)))
	return(t1);
    }
    return((struct term *) NULL);
  }
}  /* first_nonvar_arg_term */

/*************
 *
 *   first_nonvar_arg()
 *
 *   Given a clause, return the first subterm that satisfies
 *   the preceding routine first_nonvar_arg_term().
 *
 *************/

static struct term *first_nonvar_arg(struct clause *c)
{
  struct literal *lit;
  struct term *t;

  for (lit = c->first_lit; lit; lit = lit->next_lit) {
    if ((t = first_nonvar_arg_term(lit->atom, 0)))
      return(t);
  }
  return((struct term *) NULL);
}  /* first_nonvar_arg */

/*************
 *
 *   flatten_clause()
 *
 *************/

void flatten_clause(struct clause *d)
{
  struct term *t, *t1, *t2;
  struct literal *lit;
  int vnum = MAX_VARS;

  t = first_nonvar_arg(d);
  while (t) {
    lit = get_literal();
    lit->next_lit = d->first_lit;
    d->first_lit = lit;
    lit->sign = 0;
    t1 = copy_term(t);
    t2 = get_term();
    t2->type = VARIABLE;
    t2->varnum = ++vnum;
    lit->atom = build_binary_term(Dp_sn, t1, t2);
    for (lit = lit->next_lit; lit; lit = lit->next_lit)
      lit->atom = replace_term(lit->atom, t, t2);
    t = first_nonvar_arg(d);
  }
    
}  /* flatten_clause */

#else  /* new flatten */

/*************
 *
 *   yankable_term()
 *
 *************/

static struct term *yankable_term(struct term *t, int check)
{
  if (check && t->type != VARIABLE && !domain_element(t))
    return t;
  else {
    struct rel *r;
    struct term *t1;
    for (r = t->farg; r; r = r->narg) {
      t1 = yankable_term(r->argval, 1);
      if (t1 != NULL)
	return t1;
    }
    return NULL;
  }
}  /* yankable_term */

/*************
 *
 *   yankable()
 *
 *   Given a clause, return the first yankable term.
 *
 *   A term is yankable if it is not a variable, domain element,
 *   or the left side of an equality.
 *
 *************/

static struct term *yankable(struct clause *c)
{
  struct literal *lit;
  struct term *t = NULL;

  for (lit = c->first_lit; lit; lit = lit->next_lit) {
    struct term *a = lit->atom;
    if (!is_eq(a->sym_num))
      t = yankable_term(a, 0);
    else {
      t = yankable_term(a->farg->argval, 0);
      if (t == NULL)
	t = yankable_term(a->farg->narg->argval, 1);
    }
    if (t != NULL)
      return t;
  }
  return NULL;
}  /* yankable */

/*************
 *
 *   eq_sn() -- return a sym_num for equality
 *
 *************/

static int eq_sn()
{
  if (Flags[TPTP_EQ].val)
    return str_to_sn("equal", 2);
  else
    return str_to_sn("=", 2);
}  /* eq_sn */

/*************
 *
 *   flatten_clause()
 *
 *   Apply this kind of transformation
 *
 *     p(a)    ========>      a != x | p(x)
 *
 *   to all terms except domain_elements and left sides of equalities (+ or -)
 *
 *************/

void flatten_clause(struct clause *d)
{
  struct term *t, *t1, *t2;
  struct literal *lit;
  int vnum = MAX_VARS;

  t = yankable(d);
  while (t) {
    lit = get_literal();
    lit->next_lit = d->first_lit;
    d->first_lit = lit;
    lit->sign = 0;
    t1 = copy_term(t);
    t2 = get_term();
    t2->type = VARIABLE;
    t2->varnum = ++vnum;
    lit->atom = build_binary_term(eq_sn(), t1, t2);
    for (lit = lit->next_lit; lit; lit = lit->next_lit)
      lit->atom = replace_term(lit->atom, t, t2);
    t = yankable(d);
  }
}  /* flatten_clause */

/*************
 *
 *   check_transformed_clause()
 *
 *   This checks that the transformed clause is equality-equivalent
 *   to the original.
 *
 *************/

void check_transformed_clause(struct clause *c,
			      struct clause *orig)
{
  struct clause *d = cl_copy(c);

  process_negative_equalities(d);
  if (!kludgey_e_subsume(d, orig) || !kludgey_e_subsume(orig, d)) {
    printf("%% WARNING: possible error in check_transformed_clause:\n");
    printf("%%    original:    "); p_clause(orig);
    printf("%%    transformed: "); p_clause(d);
  }
}  /* check_transformed_clause */

#endif  /* old or new flatten */
